import ROOT
import zfit
import numpy

from normalizer import normalizer
from rk_model   import rk_model
from np_reader  import np_reader as np_rdr

import utils_noroot      as utnr
import zutils.utils      as zut
import matplotlib.pyplot as plt
import rk.utilities      as rkut
import math
import pytest
import pprint
import os

#-----------------------------
def delete_all_pars():
    d_par = zfit.Parameter._existing_params
    l_key = list(d_par.keys())

    for key in l_key:
        del(d_par[key])
#----------------------------------------------------
def rename_keys(d_data, use_txs=True):
    d_rename = {}
    if use_txs:
        d_rename[  'r1_TOS'] = d_data['d1']
        d_rename[  'r1_TIS'] = d_data['d1']

        d_rename['r2p1_TOS'] = d_data['d2']
        d_rename['r2p1_TIS'] = d_data['d2']

        d_rename['2017_TOS'] = d_data['d3']
        d_rename['2017_TIS'] = d_data['d3']

        d_rename['2018_TOS'] = d_data['d4']
        d_rename['2018_TIS'] = d_data['d4']
    else:
        d_rename[  'r1']     = d_data['d1']
        d_rename['r2p1']     = d_data['d2']
        d_rename['2017']     = d_data['d3']
        d_rename['2018']     = d_data['d4']

    return d_rename
#-----------------------------
def skip_test():
    try:
        uname = os.environ['USER']
    except:
        pytest.skip()

    if uname in ['angelc', 'campoverde']:
        return

    pytest.skip()
#----------------------
def get_const(d_val, d_var):
    d_err = { name : math.sqrt(var) for name, var in d_var.items() }

    d_con = {}
    for name in d_val:
        val = d_val[name]
        err = d_err[name]

        d_con[name] = [val, err]

    return d_con
#----------------------
def print_models(d_mod, d_val, d_var):
    d_const = get_const(d_val, d_var)

    dir_path = 'tests/normalizer/simple'
    os.makedirs(dir_path, exist_ok=True)
    for name, (model_mm, model_ee) in d_mod.items():
        zut.print_pdf(model_mm, txt_path=f'{dir_path}/{name}_mm.txt', d_const=d_const)
        zut.print_pdf(model_ee, txt_path=f'{dir_path}/{name}_ee.txt', d_const=d_const)
#----------------------
def get_model(dset):
    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache    = True
    d_eff        = rdr.get_eff()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    mod         = rk_model(preffix='simple', d_eff=d_eff, d_nent=d_rare_yld, l_dset=[dset])
    d_mod       = mod.get_model()
    d_val, d_var= mod.get_cons() 

    return d_mod, d_val, d_var
#----------------------
def test_simple():
    dset = '2018_TOS'
    d_mod, d_val, d_var = get_model(dset)
    mod_mm, mod_ee      = d_mod[dset]

    obj = normalizer(dset='2018', trig='ETOS', model=mod_ee, d_val=d_val, d_var=d_var)
    obj.out_dir = 'tests/normalizer/simple'
    res = obj.get_fit_result()

    delete_all_pars()
#----------------------
def test_all_dset():
    dset = 'all_TOS'
    d_mod, d_val, d_var = get_model(dset)
    mod_mm, mod_ee      = d_mod[dset]

    obj = normalizer(dset='all', trig='ETOS', model=mod_ee, d_val=d_val, d_var=d_var)
    obj.out_dir = 'tests/normalizer/all'
    res = obj.get_fit_result()

    delete_all_pars()
#----------------------
def main():
    test_simple()
    test_all_dset()
#----------------------
if __name__ == '__main__':
    main()




import numpy
import zfit
import math
import re
import os
import pprint
import matplotlib.pyplot as plt

from logzero     import logger  as log
from scipy.stats import poisson

#----------------------------------------------------
class model:
    def __init__(self, rk=1, preffix='', d_eff=None, d_nent=None, l_dset=None):
        '''
        Parameters
        -----------------------
        rk     (float): Value used in model, will impact toy data produced 
        preffix(str)  : Used to name parameters in case multiple models used
        d_eff  (dict): Stores full efficiency (selection + reco + acceptance) for each dataset processed
        d_nent (dict): Stores number of B -> llK entries, before selection, e.g. {r1 : 23235}
        l_dset (list): List of datasets for which the models should be created, if None, do it for all datasets i.e. r1_TOS, r2p1_TOS...
        '''
        self._obs    = zfit.Space('x', limits=(4800, 6000))
        self._rk     = rk
        self._preffix= preffix
        self._d_eff  = d_eff 
        self._d_nent = d_nent
        self._l_sub_dset = l_dset

        zfit.settings.changed_warnings.hesse_name = False

        self._l_all_dset  = ['r1', 'r2p1', '2017', '2018'] 
        self._skip_signal = None
        self._l_dset      = None
        self._d_mod       = None
        self._out_dir     = None
        self._initialized = False
    #----------------------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        if self._d_eff is None:
            log.error(f'No efficiencies found, cannot provide data')
            raise

        if self._d_nent is None:
            log.error(f'No B-> Jpsi K yields found')
            raise

        self._l_dset = self._d_eff.keys() if self._l_sub_dset is None else self._l_sub_dset
        self._cache_model()

        self._initialized = True
    #----------------------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot create: {value}')
            raise

        self._out_dir = value
        log.debug(f'Using output directory: {self._out_dir}')
    #----------------------------------------------------
    def _get_ds_model(self, ds, nent_mm, nent_ee):
        self._pdf_mm = self._get_pdf(preffix=f'mm_{ds}', nent=nent_mm)
        self._pdf_ee = self._get_pdf(preffix=f'ee_{ds}', nent=nent_ee)
    
        return self._pdf_mm, self._pdf_ee
    #----------------------------------------------------
    def _plot_data(self, key, dat):
        if self._out_dir is None:
            return

        plt_dir = f'{self._out_dir}/plots/data'
        os.makedirs(plt_dir, exist_ok=True)

        arr_dat = dat.value().numpy()

        plt.hist(arr_dat, bins=50)

        log.info(f'Saving to: {plt_dir}/{key}.png')
        plt.title(f'{key}; {dat.name}')
        plt.savefig(f'{plt_dir}/{key}.png')
        plt.close('all')
    #----------------------------------------------------
    def _get_ds_sel_nent(self, dset):
        [ds_only, trig]= dset.split('_')
        nent           = self._d_nent[ds_only]

        eff_mm, eff_ee = self._d_eff[dset]
        nent_mm        = self._rk * nent * eff_mm
        nent_ee        =            nent * eff_ee 

        log.debug(f'{dset}:')
        log.debug(f'{nent_mm:15.0f}={self._rk:5.3e} * {nent:5.0f} * {eff_mm:5.3e}')
        log.debug(f'{nent_ee:15.0f}={       1:5.3e} * {nent:5.0f} * {eff_ee:5.3e}')
        log.debug('')

        return nent_mm, nent_ee
    #----------------------------------------------------
    def _get_selected_entries(self):
        d_nent_sel = {}

        for ds in self._l_dset:
            nent_mm = 0
            nent_ee = 0
            if ds in ['all_TOS', 'all_TIS']:
                for single_ds in [ ds.replace('all', dset) for dset in self._l_all_dset ]:
                    nmm, nee = self._get_ds_sel_nent(single_ds)
                    nent_mm += nmm
                    nent_ee += nee
            else:
                nent_mm, nent_ee = self._get_ds_sel_nent(ds)

            d_nent_sel[ds] = nent_mm, nent_ee

        return d_nent_sel
    #----------------------------------------------------
    def _cache_model(self):
        d_nent_sel = self._get_selected_entries()
        d_mod      = {}

        mod_mm_tos = None
        for ds, (nent_mm, nent_ee) in d_nent_sel.items():
            if self._l_sub_dset is not None and ds not in self._l_sub_dset:
                continue

            mod_mm, mod_ee = self._get_ds_model(ds, nent_mm, nent_ee)
            if 'TIS' in ds:
                mod_mm = mod_mm_tos
            else:
                mod_mm_tos = mod_mm

            d_mod[ds]      = mod_mm, mod_ee

            self._plot_model(f'{ds}_mm', mod_mm)
            self._plot_model(f'{ds}_ee', mod_ee)
    
        self._d_mod = d_mod
    #----------------------------------------------------
    def get_model(self, skip_signal=False):
        '''
        Returns
        -----------------
        d_ent (dict): Returns a dictionary: {name : tup} where
            name (str)  : model identifier, e.g. r1_ETOS
            tup  (tuple): Tuple with muon and electron PDFs, e.g. pdf_mm, pdf_ee 

        For muon, TIS model is the TOS one.
        '''
        self._skip_signal = skip_signal

        self._initialize()

        return self._d_mod
    #----------------------------------------------------
    def _add_ck(self, d_par):
        regex='nsg_ee_([0-9a-z]+_(TIS|TOS)_.*)'
        d_par_ck = {}
        for var_name in d_par:
            mtch = re.match(regex, var_name)
            if not mtch:
                continue

            suffix = mtch.group(1)

            ee_yld_name = var_name
            mm_yld_name = var_name.replace('_ee_', '_mm_').replace('_TIS_', '_TOS_')

            ee_yld, _   = d_par[ee_yld_name]
            mm_yld, _   = d_par[mm_yld_name]

            d_par_ck[f'ck_{suffix}'] = [ (self._rk * ee_yld) / mm_yld, 0]

        d_par.update(d_par_ck)

        return d_par
    #----------------------------------------------------
    def _add_ext_constraints(self, d_par, d_var):
        if d_var is None:
            log.warning(f'Not adding errors for constrained parameters in prefit dictionary')
            return d_par

        d_par_new = {}
        for name, var in d_var.items():
            if name not in d_par:
                log.error(f'Cannot find {name} in prefit dictionary:')
                pprint.pprint(d_par.keys())
                raise

            val = d_par[name][0]
            err = math.sqrt(var)

            d_par_new[name] = [val, err]

        d_par.update(d_par_new)

        return d_par
    #----------------------------------------------------
    def _add_ck_constraints(self, d_par, ck_cov):
        if ck_cov is None:
            log.warning(f'Not adding errors for ck parameters in prefit dictionary')
            return d_par

        d_par_new={}
        counter=0
        for name, [val, _] in d_par.items():
            if not name.startswith('ck_'):
                continue

            var = ck_cov[counter][counter]
            err = math.sqrt(var)

            d_par_new[name] = [val, err]
            counter+=1

        d_par.update(d_par_new)

        return d_par
    #----------------------------------------------------
    def get_prefit_pars(self, d_var=None, ck_cov=None):
        '''
        Used to get model parameters used to make the toy data

        Parameters
        --------------------
        d_var (dict): Dictionary with variances for parameters that are constrained. If pased the
        constraint widths will be added as errors in the prefit dictionary

        ck_cov (numpy.array): nxn numpy array representing covariance matrix for ck parameters

        Returns 
        --------------------
        d_par (dict): Dictionary storing the prefit parameters (used to build the model) and their
        errors, e.g. {'par_x' : (3, 1)}
        '''
        self._initialize()

        d_model = self.get_model()

        d_par = {}
        for mod_mm, mod_ee in d_model.values():
            d_par_mm = { par.name : [ par.value().numpy(), 0] for par in mod_mm.get_params() }
            d_par_ee = { par.name : [ par.value().numpy(), 0] for par in mod_ee.get_params() }

            d_par.update(d_par_ee)
            d_par.update(d_par_mm)

        d_par['rk'] = [self._rk, 0]
        d_par       = self._add_ck(d_par)
        d_par       = self._add_ext_constraints(d_par, d_var)
        d_par       = self._add_ck_constraints(d_par, ck_cov)

        return d_par
    #----------------------------------------------------
    def get_data(self, rseed=3):
        '''
        Returns toy data from model

        Parameters:
        -----------------
        rseed  (int):  Random seed

        Returns:
        -----------------
        d_data (dict): Dictionary with dataset and tuple of zfit data objects paired, i.e. {r1_TOS : (pdf_mm, pdf_ee) }

        For muon, TIS dataset is the TOS one.
        '''
        self._initialize()

        zfit.settings.set_seed(rseed)

        d_data     = {}
        dst_mm_tos = None
        for ds, (pdf_mm, pdf_ee) in self._d_mod.items():
            dst_mm         = pdf_mm.create_sampler(fixed_params=False)
            dst_ee         = pdf_ee.create_sampler(fixed_params=False)

            if 'TIS' in ds:
                dst_mm = dst_mm_tos
            else:
                dst_mm_tos = dst_mm

            d_data[ds]     = dst_mm, dst_ee

            self._plot_data(f'{ds}_mm', dst_mm)
            self._plot_data(f'{ds}_ee', dst_ee)

            log.debug(f'Electron data: {dst_ee.numpy().shape[0]}')
            log.debug(f'Muon data: {dst_mm.numpy().shape[0]}')

        return d_data
    #----------------------------------------------------
    @staticmethod
    def show(d_mod):
        s_dset = { key.split('_')[0] for key in d_mod }
        for dset in s_dset:
            pdf_mm_tos, pdf_ee_tos = d_mod[f'{dset}_TOS']
            pdf_mm_tis, pdf_ee_tis = d_mod[f'{dset}_TIS']

            l_par_name_mm_tos = ', '.join([ par.name for par in pdf_mm_tos.get_params() ])
            l_par_name_mm_tis = ', '.join([ par.name for par in pdf_mm_tis.get_params() ])
            l_par_name_ee_tos = ', '.join([ par.name for par in pdf_ee_tos.get_params() ])
            l_par_name_ee_tis = ', '.join([ par.name for par in pdf_ee_tis.get_params() ])

            log.info('')
            log.info(f'{dset}')
            log.info('-' * 20)
            log.info(f'{"mm TOS":<10}{l_par_name_mm_tos:<60}')
            log.info(f'{"mm TIS":<10}{l_par_name_mm_tis:<60}')
            log.info(f'{"ee TOS":<10}{l_par_name_ee_tos:<60}')
            log.info(f'{"ee TIS":<10}{l_par_name_ee_tis:<60}')
            log.info('-' * 20)
    #----------------------------------------------------
    @staticmethod
    def get_cov(kind='diag_eq', c = 0.01):
        if   kind == 'diag_eq':
            mat = numpy.diag([c] * 8)
        elif kind == 'random':
            mat = numpy.random.rand(8, 8)
            numpy.fill_diagonal(mat, 1)
            mat = mat * c
        else:
            log.error(f'Invalid kind: {kind}')
            raise
    
        return mat 
    #----------------------------------------------------
    @staticmethod
    def get_rjpsi(kind='one'):
        d_rjpsi = {}
    
        if   kind == 'one':
            d_rjpsi['d1'] = 1 
            d_rjpsi['d2'] = 1 
            d_rjpsi['d3'] = 1 
            d_rjpsi['d4'] = 1 
        elif kind == 'eff_bias':
            d_rjpsi['d1'] = 0.83333333 
            d_rjpsi['d2'] = 0.83333333 
            d_rjpsi['d3'] = 0.83333333 
            d_rjpsi['d4'] = 0.83333333 
        else:
            log.error(f'Wrong kind: {kind}')
            raise
    
        return d_rjpsi
    #----------------------------------------------------
    @staticmethod
    def get_eff(kind='equal'):
        d_eff = {}
        if   kind == 'diff':
            d_eff['d1'] = (0.6, 0.3)
            d_eff['d2'] = (0.5, 0.2)
            d_eff['d3'] = (0.7, 0.3)
            d_eff['d4'] = (0.8, 0.4)
        elif kind == 'half':
            d_eff['d1'] = (0.6, 0.3)
            d_eff['d2'] = (0.6, 0.3)
            d_eff['d3'] = (0.6, 0.3)
            d_eff['d4'] = (0.6, 0.3)
        elif kind == 'equal':
            d_eff['d1'] = (0.3, 0.3)
            d_eff['d2'] = (0.3, 0.3)
            d_eff['d3'] = (0.3, 0.3)
            d_eff['d4'] = (0.3, 0.3)
        elif kind == 'bias':
            d_eff['d1'] = (0.6, 0.25)
            d_eff['d2'] = (0.6, 0.25)
            d_eff['d3'] = (0.6, 0.25)
            d_eff['d4'] = (0.6, 0.25)
        else:
            log.error(f'Invalid kind: {kind}')
            raise
    
        return d_eff
#----------------------------------------------------




from logzero import logger as log

import os
import glob
import ROOT
import math
import numpy
import utils_noroot      as utnr
import zutils.utils      as zut
import matplotlib.pyplot as plt

from zutils.plot   import plot     as zfp
from fitter        import zfitter

#--------------------------------------
class normalizer:
    '''
    Class used to find normalizations for combinatorial and PRec components
    in the high-q2 signal model
    '''
    #--------------------------------------
    def __init__(self, dset=None, trig=None, model=None, d_val=None, d_var=None):
        self._dset    = dset
        self._trig    = trig
        self._model   = model 
        self._d_val   = d_val
        self._d_var   = d_var

        self._l_flt_par= ['lm_cb', 'mu_cb', 'ncb', 'npr_ee']
        self._dat_path = f'{os.environ["CASDIR"]}/tools/apply_selection/blind_fits/data/v10.21p2'

        self._d_const  = {}
        self._nbins    = 60
        self._hig_mm   = 5360, 5600
        self._low_mm   = 5180, 5200 
        self._hig_ee   = 5450, 5950 
        self._low_ee   = 4450, 5000 

        self._out_dir     = None
        self._initialized = False 
    #--------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        os.makedirs(f'{self._out_dir}/pdf', exist_ok=True)
        self._d_const= self._prepare_model(self._model)

        zut.print_pdf(self._model, txt_path=f'{self._out_dir}/pdf/{self._dset}_{self._trig}.txt', d_const=self._d_const)

        self._initialized = True 
    #--------------------------------------
    def _get_data_paths(self):
        if   self._dset == 'r1':
            l_dset = [f'{self._dat_path}/2011_{self._trig}', f'{self._dat_path}/2012_{self._trig}']
        elif self._dset == 'r2p1':
            l_dset = [f'{self._dat_path}/2015_{self._trig}', f'{self._dat_path}/2016_{self._trig}']
        elif self._dset == 'all':
            l_dset = [f'{self._dat_path}/{year}_{self._trig}' for year in [2011, 2012, 2015, 2016, 2017, 2018] ]
        else:
            l_dset = [f'{self._dat_path}/{self._dset}_{self._trig}']

        l_wc = [f'{dset}/*.root' for dset in l_dset]

        l_l_path = [ glob.glob(wc) for wc in l_wc ]
        for l_path, wc in zip(l_l_path, l_wc):
            if len(l_path) == 0:
                log.error(f'No file found in {wc}')
                raise

        return l_wc
    #-------------------------------------
    def _get_mass(self, l_dpath):
        rdf = ROOT.RDataFrame(self._trig, l_dpath)
        arr_mass = rdf.AsNumpy(['B_M'])['B_M']

        low = self._low_mm if self._trig == 'MTOS' else self._low_ee
        hig = self._hig_mm if self._trig == 'MTOS' else self._hig_ee

        arr_mass_lo = arr_mass[(arr_mass < low[1])]
        arr_mass_hi = arr_mass[(arr_mass > hig[0])]

        arr_mass_cn = arr_mass[(arr_mass > low[1]) & (arr_mass < hig[0])]
        arr_mass_cn = numpy.random.uniform(low=low[1], high=hig[0], size=arr_mass_cn.shape)

        arr_mass = numpy.concatenate([arr_mass_lo, arr_mass_cn, arr_mass_hi])

        return arr_mass
    #--------------------------------------
    def _get_data(self):
        dat_path = f'{self._out_dir}/data/{self._dset}_{self._trig}.json'
        if os.path.isfile(dat_path):
            log.info(f'Loading cached data from: {dat_path}')
            l_dat = utnr.load_json(dat_path)
            return numpy.array(l_dat)

        os.makedirs(f'{self._out_dir}/data', exist_ok=True)

        l_dpath = self._get_data_paths()
        arr_dat = self._get_mass(l_dpath)

        log.info(f'Saving to: {dat_path}')
        utnr.dump_json(arr_dat.tolist(), dat_path) 
        self._plot_data(arr_dat)

        return arr_dat
    #--------------------------------------
    def _plot_data(self, arr_mass):
        if not self._out_dir:
            return

        rng_ee = 4800, 6000
        rng_mm = 5100, 5600
        rng = rng_mm if self._trig == 'MTOS' else rng_ee
        plt.hist(arr_mass, range=rng, bins=50, histtype='step')
        plot_path = f'{self._out_dir}/data/{self._dset}_{self._trig}.png'
        log.info(f'Saving to: {plot_path}')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------------------
    def _prepare_model(self, pdf):
        s_par = pdf.get_params()
        for par in s_par:
            par.floating = False

        d_const = {}
        for par in s_par:
            for flt_par in self._l_flt_par:
                if par.name.startswith(flt_par):
                    par.floating = True
                else:
                    continue

                if par.name in self._d_val:
                    var = self._d_var[par.name]
                    val = self._d_val[par.name]
                    d_const[par.name] = val, math.sqrt(var)

        return d_const
    #--------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot create directory: {value}')
            raise

        self._out_dir = value
    #--------------------------------------
    def _plot_fit(self, data=None, name=None, result=None, l_range=None, stacked=None):
        obj= zfp(data=data, model=self._model, result=result)
        obj.plot(nbins=self._nbins, ranges=l_range, stacked=stacked)

        if   '_all_' not in name and 'MTOS' in name:
            obj.axs[0].set_ylim(0, 130)
        elif '_all_' not in name and 'ETOS' in name:
            obj.axs[0].set_ylim(0,  70)
        elif '_all_' not in name and 'GTIS' in name:
            obj.axs[0].set_ylim(0,  20)

        obj.axs[1].set_ylim(-5, 5)
        obj.axs[1].axhline(0, color='r')

        os.makedirs(f'{self._out_dir}/fits', exist_ok=True)
        plot_path = f'{self._out_dir}/fits/{name}.png'
        log.info(f'Saving to: {plot_path}')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------------------
    def get_fit_result(self):
        self._initialize()

        arr_mass= self._get_data()

        low   = self._low_mm if self._trig == 'MTOS' else self._low_ee
        hig   = self._hig_mm if self._trig == 'MTOS' else self._hig_ee

        obj   = zfitter(self._model, arr_mass)
        res   = obj.fit(ranges=[low, hig], d_const=self._d_const)

        self._plot_fit(data=arr_mass, name=f'{self._trig}_{self._dset}_stk', result=res, l_range=[low, hig], stacked= True)
        self._plot_fit(data=arr_mass, name=f'{self._trig}_{self._dset}_ovr', result=res, l_range=[low, hig], stacked=False)

        res.hesse()
        res.freeze()
        print(res)

        return res
#--------------------------------------




from logzero           import logger         as log
import os
import numpy
import pprint
import tarfile
import jacobi       as jac
import utils_noroot as utnr

from stats import average as stav

#----------------------------------------
class cs_reader:
    '''
    Class used to MC version of parameters used in rare fit model
    '''
    #------------------------
    def __init__(self, version=None, preffix=None): 
        '''
        version (str): Version of fit to rare mode 
        '''
        self._ver         = version 
        self._preffix     = preffix 

        self._l_ds_lab    = ['r1_TOS', 'r1_TIS', 'r2p1_TOS', 'r2p1_TIS', '2017_TOS', '2017_TIS', '2018_TOS', '2018_TIS']
        self._l_trig      = ['MTOS', 'ETOS', 'GTIS']
        self._l_year      = ['2011', '2012', '2015', '2016', '2017', '2018']
        self._l_dset      = ['r1', 'r2p1', '2017', '2018']
        self._l_brem      = ['0', '1', '2']
        self._l_params    = ['mu', 'sg']
        self._d_val       = {}
        self._d_var       = {}

        self._d_d_d_mc    = None 
        self._d_d_d_dt    = None 
        self._cache_dir   = None
        self._cache       = False
        self._initialized = False 
    #------------------------
    @property
    def cache(self):
        return self._cache

    @cache.setter
    def cache(self, value):
        if value not in [True, False]:
            log.error(f'Invalid cache value: {value}')
            raise ValueError

        self._cache = value
    #------------------------
    @property
    def cache_dir(self):
        if self._cache_dir is None:
            return

        return cache_dir

    @cache_dir.setter
    def cache_dir(self, value):
        cache_dir = f'{value}'
        try:
            os.makedirs(cache_dir, exist_ok=True)
        except:
            log.error(f'Cannot create: {cache_dir}')
            raise ValueError

        self._cache_dir = cache_dir
    #------------------------
    def _initialize(self):
        if self._initialized:
            return

        if self._cache_dir is None:
            self.cache_dir = '/tmp/cs_reader'
            log.warning(f'Using default caching directory: {self._cache_dir}')

        self._set_paths()
        self._cache_data('signal')
        self._cache_data('data')
        self._tar_data()

        self._initialized = True
    #------------------------
    def _set_paths(self):
        if self._cache_dir is not None:
            log.info(f'Cache dir: {self._cache_dir}')
        else: 
            self._cache_dir = f'/tmp/cs_reader/{self._vers}'
            log.debug(f'Cache dir: {self._cache_dir}')

        os.makedirs(self._cache_dir, exist_ok=True)
    #------------------------
    def _merge_brem(self, d_d_par_brem):
        d_par = {}
        for param in self._l_params:
            l_val = []
            l_err = []
            for brem in self._l_brem:
                val, err  = d_d_par_brem[brem][param]
                l_val.append(val)
                l_err.append(err)

            arr_val = numpy.array(l_val)
            arr_err = numpy.array(l_err)

            avg, err, pval = stav.average(arr_val, arr_err)

            d_par[param] = avg, err

        return d_par
    #------------------------
    def _merge_runs(self, d_d_par_year):
        d_r1   = self._merge_years('2011', '2012', d_d_par_year)
        d_r2p1 = self._merge_years('2015', '2016', d_d_par_year)

        d_d_par_dset         = {}
        d_d_par_dset['r1']   = d_r1
        d_d_par_dset['r2p1'] = d_r2p1
        d_d_par_dset['2017'] = d_d_par_year['2017']
        d_d_par_dset['2018'] = d_d_par_year['2018']

        return d_d_par_dset
    #------------------------
    def _merge_years(self, y1, y2, d_d_par_year):
        d_par_y1 = d_d_par_year[y1]
        d_par_y2 = d_d_par_year[y2]
        d_par_ds = {}
        for param in self._l_params:
            val_y1, err_y1 = d_par_y1[param]
            val_y2, err_y2 = d_par_y2[param]

            arr_val = numpy.array([val_y1, val_y2])
            arr_err = numpy.array([err_y1, err_y2])

            avg, err, pval = stav.average(arr_val, arr_err)

            d_par_ds[param] =  avg, err

        return d_par_ds
    #------------------------
    def _is_cached(self, json_path, name):
        if self._cache and os.path.isfile(json_path):
            log.info(f'Loading cached parameters from: {json_path}')

            if name == 'signal':
                self._d_d_d_mc = utnr.load_json(json_path)
            else:
                self._d_d_d_dt = utnr.load_json(json_path)

            return True

        return False
    #------------------------
    def _cache_data(self, name):
        json_path = f'{self._cache_dir}/constraints_{name}.json'
        if self._is_cached(json_path, name):
            return

        log.info('Reloading parameters')
        d_d_d_par_year_trig = {} 
        for trig in self._l_trig:
            d_d_par_year = {}
            for year in self._l_year:
                d_d_par_brem = {}
                l_brem = ['0'] if trig == 'MTOS' else self._l_brem
                for brem in l_brem:
                    d_par = self._get_pars(trig, year, brem, name)
                    d_d_par_brem[brem] = d_par

                d_d_par_year[year] = d_d_par_brem['0'] if trig == 'MTOS' else self._merge_brem(d_d_par_brem)

            d_d_d_par_year_trig[trig] = self._merge_runs(d_d_par_year)

        d_val = self._format_data(d_d_d_par_year_trig)
        if name == 'signal':
            self._d_d_d_mc = d_val 
        else:
            self._d_d_d_dt = d_val 

        log.info(f'Saving to: {json_path}')
        utnr.dump_json(d_val, json_path)
    #------------------------
    def _tar_data(self):
        tar_path = f'{self._cache_dir}.tar.gz'

        if os.path.isfile(tar_path):
            return

        json_dir = os.path.basename(self._cache_dir)
        with tarfile.open(tar_path, 'w:gz') as tar:
            tar.add(self._cache_dir, arcname=json_dir)

        log.info(f'Making: {tar_path}')
    #------------------------
    def _format_data(self, d_d_d_par_year_trig):
        d_d_d_dset_parm_trig = {}
        for dset in self._l_dset:
            d_d_parm_trig= {}
            for param in self._l_params:
                d_trig = {}
                v_mtos, e_mtos = d_d_d_par_year_trig['MTOS'][dset][param]
                v_etos, e_etos = d_d_d_par_year_trig['ETOS'][dset][param]
                v_gtis, e_gtis = d_d_d_par_year_trig['GTIS'][dset][param]

                d_trig['MTOS'] = v_mtos, e_mtos
                d_trig['ETOS'] = v_etos, e_etos
                d_trig['GTIS'] = v_gtis, e_gtis

                d_d_parm_trig[param] = d_trig

            d_d_d_dset_parm_trig[dset] = d_d_parm_trig

        return d_d_d_dset_parm_trig
    #------------------------
    def _get_pars(self, trig, year, brem, name):
        cas_dir = os.environ['CASDIR']

        jsn_path= f'{cas_dir}/monitor/mass_scales/{self._ver}/{year}_{trig}/pars/cat_{brem}/{name}.json'
        d_data  = utnr.load_json(jsn_path)

        d_par       = {}
        d_par['mu'] = d_data['mu']
        d_par['sg'] = d_data['sg']

        return d_par
    #------------------------
    def _compute_const(self):
        for trig in self._l_trig:
            for dset in self._l_dset:
                for par in self._l_params:
                    val_mc = self._d_d_d_mc[dset][par][trig]
                    val_dt = self._d_d_d_dt[dset][par][trig]


                    val, var         = self._get_val(par, val_mc, val_dt)
                    key              = self._get_key(par, trig, dset)
                    self._d_val[key] = val
                    self._d_var[key] = var
    #------------------------
    def _get_key(self, par, trig, dset):
        chan = 'mm'  if trig == 'MTOS' else 'ee'
        trig = 'TIS' if trig == 'GTIS' else 'TOS'
        par  = {'mu' : 'dmu', 'sg' : 'rsg'}[par]

        return f'{par}_{chan}_{dset}_{trig}_{self._preffix}'
    #------------------------
    def _get_val(self, name, val_mc, val_dt):
        if name == 'mu':
            val, var = jac.propagate(lambda x : x[0] - x[1], [val_dt[0], val_mc[0]], [[val_dt[1] ** 2, 0], [0, val_mc[1] ** 2]])
        elif name == 'sg':
            val, var = jac.propagate(lambda x : x[0] / x[1], [val_dt[0], val_mc[0]], [[val_dt[1] ** 2, 0], [0, val_mc[1] ** 2]])
        else:
            log.error(f'Invalid parameter: {name}')
            raise ValueError

        return float(val), float(var)
    #------------------------
    def get_constraints(self):
        '''
        Will return a dictionary with values in MC for a given fitting parameter

        Returns
        ------------------
        x, y (tuple) : Tuple with x and y where

        x: Dictionary mapping parameter name and value to which to constrain it {name : val} 
        x: Dictionary mapping parameter name and variance to which to constrain it {name : var} 
        '''
        self._initialize()

        self._compute_const()

        return self._d_val, self._d_var
#----------------------------------------




import ROOT
import zfit
import numpy

from rk_model   import rk_model
from mc_reader  import mc_reader as mc_rdr
from np_reader  import np_reader as np_rdr
from cs_reader  import cs_reader as cs_rdr

import matplotlib.pyplot as plt
import rk.utilities      as rkut
import pytest
import pprint
import os

#-----------------------------
def delete_all_pars():
    d_par = zfit.Parameter._existing_params
    l_key = list(d_par.keys())

    for key in l_key:
        del(d_par[key])
#----------------------------------------------------
def rename_keys(d_data, use_txs=True):
    d_rename = {}
    if use_txs:
        d_rename[  'r1_TOS'] = d_data['d1']
        d_rename[  'r1_TIS'] = d_data['d1']

        d_rename['r2p1_TOS'] = d_data['d2']
        d_rename['r2p1_TIS'] = d_data['d2']

        d_rename['2017_TOS'] = d_data['d3']
        d_rename['2017_TIS'] = d_data['d3']

        d_rename['2018_TOS'] = d_data['d4']
        d_rename['2018_TIS'] = d_data['d4']
    else:
        d_rename[  'r1']     = d_data['d1']
        d_rename['r2p1']     = d_data['d2']
        d_rename['2017']     = d_data['d3']
        d_rename['2018']     = d_data['d4']

    return d_rename
#-----------------------------
def skip_test():
    try:
        uname = os.environ['USER']
    except:
        pytest.skip()

    if uname in ['angelc', 'campoverde']:
        return

    pytest.skip()
#----------------------
def test_simple():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='simple', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    mod.out_dir = 'tests/rk_model/simple'
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
def test_all_years():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='allyears', d_eff=d_eff, d_nent=d_nent, l_dset=['all_TOS'])
    mod.out_dir = 'tests/rk_model/all_years'
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
def test_data():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='data', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    d_mod       = mod.get_data()

    delete_all_pars()
#----------------------
def test_rseed():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod_1   = rk_model(preffix='data1', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    d_dat_1 = mod_1.get_data(rseed=0)

    mod_2   = rk_model(preffix='data2', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    d_dat_2 = mod_2.get_data(rseed=0)

    plt_dir = f'tests/rk_model/rseed'
    os.makedirs(plt_dir, exist_ok=True)
    for index in [0, 1]:
        kind = 'mm' if index == 0 else 'ee'
        for key in d_dat_1:
            arr_dat_1 = d_dat_1[key][index].numpy().flatten()
            arr_dat_2 = d_dat_2[key][index].numpy().flatten()

            close = numpy.allclose(arr_dat_1, arr_dat_2, atol=1e-5)
            assert(close)

            min_x, max_x = min(arr_dat_1), max(arr_dat_1)

            plt.hist(arr_dat_1, bins=30, range=(min_x, max_x), histtype='step', linestyle='-', label='First')
            plt.hist(arr_dat_2, bins=30, range=(min_x, max_x), histtype='step', linestyle=':', label='Second')
            plt.legend()
            plt.savefig(f'{plt_dir}/{key}_{kind}.png')
            plt.close('all')

    delete_all_pars()
#----------------------
def test_cons():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='cons', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    d_mod       = mod.get_cons()

    delete_all_pars()
#----------------------
def main():
    test_data()
    test_cons()
    test_rseed()
    test_simple()
    test_all_years()
#----------------------
if __name__ == '__main__':
    main()




from cs_reader import cs_reader as cs_rdr

import pprint

#----------------------------------------
def test_cache():
    rdr           = cs_rdr(version='v4')
    rdr.cache     = True
    rdr.cache_dir = 'tests/cs_reader/v4_csrdr'
    x, y = rdr.get_constraints()
#----------------------------------------
def test_simple():
    rdr          = cs_rdr(version='v4')
    d_val, d_var = rdr.get_constraints()

    pprint.pprint(d_val)
    pprint.pprint(d_var)
#----------------------------------------
def main():
    test_cache()
    test_simple()
#----------------------------------------
if __name__ == '__main__':
    main()




import re
import ROOT
import zfit
import math
import numpy
import pprint
import pandas            as pnd
import utils_noroot      as utnr
import matplotlib.pyplot as plt

from logzero     import logger as log
from zutils.plot import plot   as zfp

#--------------------------------
class extractor:
    '''
    Class used to extract RK from data, model, constraints, etc
    '''
    def __init__(self, dset=None, drop_correlations=False):
        '''
        Parameters
        ------------------
        dset (list): What datasets should be used? e.g. [r1p2, 2018]
        drop_correlations (bool): Will diagonalize the covariance matrix used for constraints
        if set to True, default is False.
        '''
        self._l_dset      = dset 
        self._drop_corr   = drop_correlations

        self._preffix     = None 
        self._d_eff       = None
        self._cov         = None
        self._d_rjpsi     = None
        self._d_data      = None
        self._d_model     = None
        self._d_ck        = None 
        self._d_val       = None
        self._d_var       = None
        self._d_par       = None
        self._l_fix       = None
        self._d_fix       = {'Name' : [], 'Value' : []} 
        self._l_dsname    = ['r1_TOS', 'r1_TIS', 'r2p1_TOS', 'r2p1_TIS', '2017_TOS', '2017_TIS', '2018_TOS', '2018_TIS']

        self._rk          = zfit.Parameter('rk', 1.0, 0.0, 2.0)

        self._plt_dir     = None

        self._is_combined = self._l_dset == ['all_TOS'] or self._l_dset == ['all_TIS']
        self._has_const   = False
        self._initialized = False
    #--------------------------------
    def _initialize(self):
        if self._initialized:
            return

        if self._l_dset is not None:
            self._l_dsname = self._l_dset

        self._check_inputs()
        self._d_ck = { dsname : zfit.Parameter(f'ck_{dsname}_{self._preffix}', 1.0, 0.0, 2.0) for dsname in self._l_dsname }

        d_model = {}
        for ds, (mod_mm, mod_ee) in self._d_model.items():
            mod_mm = self._reparametrize_mm_pdf(ds, mod_mm, mod_ee)

            d_model[ds] = (mod_mm, mod_ee)

        self._d_model = d_model
        self._d_par   = self._get_model_pars()

        self._check_constraints()
        self._fix_pars()

        self._initialized = True
    #--------------------------------
    def _get_model_pars(self):
        d_par = {}
        for mod_mm, mod_ee in self._d_model.values():
            d_par_mm = { par.name : par for par in mod_mm.get_params() }
            d_par_ee = { par.name : par for par in mod_ee.get_params() }

            d_par.update(d_par_mm)
            d_par.update(d_par_ee)

        return d_par
    #--------------------------------
    def _fix_pars(self):
        if self._l_fix is None:
            return

        for fix_par in self._l_fix:
            l_par_name = []
            for par_name, par in self._d_par.items():
                if par_name.startswith(fix_par):
                    l_par_name.append(par_name)
                    self._add_fixed_par(par)
            if len(l_par_name) == 0:
                log.error(f'No parameters found to fixed for: {fix_par}')
                raise
    #--------------------------------
    def _check_constraints(self):
        '''
        Checks that every parameter constrained, exists in the model
        '''
        if not self._has_const:
            return

        for var_name in self._d_val.keys():
            if var_name not in self._d_par:
                log.error(f'Constraint for missing model variable: {var_name}')
                pprint.pprint(l_mod_var_name)
                raise
    #--------------------------------
    def _check_inputs(self):
        for ds, (eff_mm, eff_ee) in self._d_eff.items():
            self._check_eff(eff_mm)
            self._check_eff(eff_ee)

        for ds, (mod_mm, mod_ee) in self._d_model.items():
            self._check_model(mod_mm)
            self._check_model(mod_ee)

        for ds, (dat_mm, dat_ee) in self._d_data.items():
            self._check_data(dat_mm)
            self._check_data(dat_ee)

        self._check_keys(self._d_eff  , self._d_model)
        self._check_keys(self._d_model, self._d_data )
        self._check_keys(self._d_data , self._d_eff  )

        self._check_cov()
        self._check_rjpsi()
    #--------------------------------
    def _check_keys(self, d1, d2):
        if d1.keys() != d2.keys():
            log.error('Keys differ:')
            log.error(d1.keys())
            log.error(d2.keys())
            raise ValueError
    #--------------------------------
    def _check_cov(self):
        if self._cov is None:
            return

        if not isinstance(self._cov, numpy.ndarray):
            log.error(f'Covariance matrix is not a numpy array')
            raise

        ndset = len(self._d_eff)

        if self._cov.shape != (ndset, ndset):
            log.error(f'Covariance matrix has wrong shape: {self._cov.shape}')
            raise
    #--------------------------------
    def _check_rjpsi(self):
        if self._d_rjpsi is None:
            self._d_rjpsi = {dsname : 1 for dsname in self._l_dsname}
            log.warning('rjpsi not found, using 1 for all datasets')
    #--------------------------------
    def _reparametrize_mm_pdf(self, dsname, mod_mm, mod_ee):
        nsg_mm  = self._get_yld(mod_mm, 'nsg_mm')
        nsg_ee  = self._get_yld(mod_ee, 'nsg_ee')

        l_model = mod_mm.models
        s_param = mod_mm.get_params(is_yield=True)
        l_param = list(s_param)
        index   = l_param.index(nsg_mm)

        eff_mm, eff_ee = self._d_eff[dsname]

        try:
            rjpsi = self._d_rjpsi[dsname]
        except:
            log.error(f'Cannot extract rjpsi for {dsname} only found:')
            log.info(self._d_rjpsi.keys())
            raise

        ck=self._d_ck[dsname]
        ck.set_value( (eff_ee / eff_mm) / rjpsi )

        if self._cov is None:
            self._add_fixed_par(ck)

        nsg_mm = zfit.ComposedParameter(f'nsg_mm_rk_{dsname}', lambda a, b, c: a * b / c, params=[self._rk, nsg_ee, ck])

        log.debug(f'nsg_mm_rk_{dsname} -> {nsg_ee.name} * {self._rk.name} / {ck.name}')

        l_param[index] = nsg_mm

        l_model_ext  = [ model.copy().create_extended(nevt) for model, nevt in zip(l_model, l_param) ]

        mod_mm = zfit.pdf.SumPDF(l_model_ext)

        return mod_mm
    #--------------------------------
    def _add_fixed_par(self, par):
        par.floating          = False
        self._d_fix['Name']  += [par.name]
        self._d_fix['Value'] += [float(par.numpy())]

        log.warning(f'{par.name:<20}{"=":20}{par.value():<20.3e}')
    #--------------------------------
    def _get_yld(self, model, preffix):
        l_param   = model.get_params(is_yield=True)

        l_sig_yld = [ param for param in l_param if param.name.startswith(preffix)]
        [sig_yld] = l_sig_yld

        return sig_yld
    #--------------------------------
    @property
    def plt_dir(self):
        return self._plt_dir 

    @plt_dir.setter
    def plt_dir(self, value):
        try:
            self._plt_dir = utnr.make_dir_path(value)
        except:
            log.error(f'Cannot create: {value}')
            raise
    #--------------------------------
    @property
    def fix(self):
        return self._l_fix

    @fix.setter
    def fix(self, value):
        self._l_fix = value
    #--------------------------------
    @property
    def rjpsi(self):
        return self._d_rjpsi

    @rjpsi.setter
    def rjpsi(self, value):
        self._d_rjpsi = self._filter_values(value)
    #--------------------------------
    def _filter_cov(self, mat):
        if   self._l_dset is None:
            log.info(f'Using all datasets, not filtering covariance')
            return mat
        elif self._is_combined:
            log.info(f'Using combined dataset, not filtering covariance')
            return mat

        log.info(f'Filtering covariance for: {self._l_dset}')
        if set(self._l_dsname) == set(self._l_dset):
            return mat

        l_dset_drop  = [ dset_drop for dset_drop in self._l_dsname if dset_drop not in self._l_dset ]
        l_index_drop = [ self._l_dsname.index(dset_drop) for dset_drop in l_dset_drop ]

        mat=numpy.delete(mat, obj=l_index_drop, axis=0)
        mat=numpy.delete(mat, obj=l_index_drop, axis=1)

        return mat
    #--------------------------------
    def _filter_const(self, value):
        if self._l_dset is None:
            return value

        d_data = {}
        for key, val in value.items():
            keep=False
            for dset in self._l_dset:
                if dset in key:
                    keep=True
                    break

            if keep:
                d_data[key] = val

        return d_data
    #--------------------------------
    def _filter_values(self, value):
        if self._l_dset is None:
            return value

        d_data = {}
        for key, val in value.items():
            if key not in self._l_dset:
                continue

            d_data[key] = val

        return d_data
    #--------------------------------
    @property
    def eff(self):
        return self._d_eff

    @eff.setter
    def eff(self, value):
        self._d_eff = self._filter_values(value)
    #--------------------------------
    @property
    def cov(self):
        return self._cov

    @cov.setter
    def cov(self, value):
        if value is None:
            return
        self._cov = self._filter_cov(value)
    #--------------------------------
    @property
    def data(self):
        return self._d_data

    @data.setter
    def data(self, value):
        self._d_data = self._filter_values(value)
    #--------------------------------
    @property
    def model(self):
        return self._d_model

    @model.setter
    def model(self, value):
        self._d_model = self._filter_values(value)
    #--------------------------------
    @property
    def const(self):
        return self._d_val, self._d_var

    @const.setter
    def const(self, value):
        '''
        It will take the constraints on extra parameters. 
        All the variances that are None (in second dictionary of tuple) correspond to fixed parameters

        Parameters:
        --------------------
        tuple : Storing two dictionaries {par_name : par_value}, {par_name : par_variance}
        '''
        self._d_val = self._filter_const(value[0])
        self._d_var = self._filter_const(value[1])

        if self._d_val.keys() != self._d_var.keys():
            log.error(f'Keys of values and variances differ')
            raise ValueError

        self._has_const = True
    #--------------------------------
    def _check_data(self, obj):
        if not isinstance(obj, zfit.data.Data):
            log.error(f'Object introduced is not a zfit dataset')
            raise ValueError
    #--------------------------------
    def _check_model(self, obj):
        if not isinstance(obj, zfit.pdf.SumPDF):
            log.error(f'Object introduced is not a zfit PDF')
            raise ValueError

        pdf = obj
        if not pdf.is_extended:
            log.error(f'PDF is not extended:')
            print(pdf)
            raise ValueError

        l_yld = pdf.get_params(is_yield=True)
        l_yld_name = [ yld.name for yld in l_yld if yld.name.startswith('nsg_') ]

        try:
            [yld_name] = l_yld_name
        except:
            log.error('Not found one and only one signal yield:')
            print(l_yld_name)
            raise ValueError

        self._check_preffix(yld_name)

        log.debug(f'Picking up component with signal yield: {yld_name}')
    #--------------------------------
    def _check_preffix(self, yld_name):
        regex= r'n[a-z]+_(ee|mm)_[a-z0-9]+_(TOS|TIS)_(.*)'
        mtch = re.match(regex, yld_name)
        if not mtch:
            log.error(f'Cannot match yield name: {yld_name} with {regex}')
            raise

        if   self._preffix is None:
            self._preffix = mtch.group(3)
        elif self._preffix != mtch.group(3):
            log.error(f'Found different prefixes: {self._preffix} -> {mtch.group(3)}')
            raise
    #--------------------------------
    def _check_eff(self, eff):
        if not isinstance(eff, float):
            log.error(f'Efficiency is not a float: {eff}')
            raise ValueError

        if not ( 0 < eff < 1 ):
            log.error(f'Efficiency is not in (0, 1): {eff:.3f}')
            raise ValueError
    #--------------------------------
    def _plot(self, data, model, results, component, stacked=None):
        if self._plt_dir is None:
            return

        obj=zfp(data=data, model=model, result=results)

        obj.plot(nbins=50, stacked=stacked) 
        obj.axs[1].set_ylim(-5, 5)
        obj.axs[1].axhline(0, color='r')

        plt_path = f'{self._plt_dir}/fit_{component}.png'
        log.info(f'Saving to: {plt_path}')
        plt.savefig(plt_path, bbox_inches='tight')
    #--------------------------------
    def _finalize(self):
        self._delete_par('rk')

        for ck in self._d_ck.values():
            self._delete_par(ck.name)

        for dsname in self._l_dsname:
            self._delete_par(f'nsg_ee_rk_{dsname}')

        df_fix=pnd.DataFrame(self._d_fix)
        df_fix.to_json(f'{self._plt_dir}/fixed_pars.json', indent=4)
    #--------------------------------
    def _delete_par(self, par_name):
        if par_name in zfit.Parameter._existing_params:
            del zfit.Parameter._existing_params[par_name]
    #--------------------------------
    def _randomize(self, l_ck_val, cov):
        l_ck_val_rdm = numpy.random.multivariate_normal(l_ck_val, cov, 1)[0]

        log.warning(f'Randomizing Gaussian mean: {l_ck_val} -> {l_ck_val_rdm}')

        return l_ck_val_rdm.tolist()
    #--------------------------------
    def _get_ck_const(self):
        if self._cov is None:
            return

        log.info('Getting CK constraints')
        l_ck_par=[]
        l_ck_val=[]
        for dsname in self._l_dsname:
            eff_mm, eff_ee = self._d_eff[dsname]
            rjpsi          = self._d_rjpsi[dsname]

            ck_val         = (eff_ee / eff_mm) / rjpsi
            ck_par         = self._d_ck[dsname]

            l_ck_val.append(ck_val)
            l_ck_par.append(ck_par)

        l_ck_val = self._randomize(l_ck_val, self._cov)

        self._print_const(l_ck_par, l_ck_val, self._cov)

        return l_ck_par, l_ck_val, self._cov
    #--------------------------------
    def _get_ex_const(self):
        if not self._has_const:
            log.info('No extra constraints found')
            return 

        log.info('Getting extra constraints')
        l_par    = [ self._d_par[parname] for parname in self._d_val.keys() ]
        l_val    = [ self._d_val[parname] for parname in self._d_val.keys() ]
        l_var    = [ self._d_var[parname] for parname in self._d_val.keys() ]
        cov      = numpy.diag(l_var)
        l_val    = self._randomize(l_val, cov)

        return l_par, l_val, cov 
    #--------------------------------
    def _print_const(self, l_par, l_val, cov):
        eq_size = len(l_par) == len(l_val) == len(cov)
        if not eq_size:
            log.error(f'Invalid sizes for constraints')
            print(l_par)
            print(l_val)
            print(cov)
            raise

        nconst = len(l_par)
        log.debug('-' * 30)
        log.debug(f'{"Parameter":<20}{"Value":>15}{"Error":>15}')
        log.debug('-' * 30)
        for i_const in range(nconst):
            par_name = l_par[i_const].name
            par_val  = l_val[i_const]
            par_err  = math.sqrt(cov[i_const][i_const])

            log.debug(f'{par_name:<20}{par_val:>15.3e}{par_err:>15.3e}')
        log.debug('-' * 30)
    #--------------------------------
    def _plot_fit(self):
        if self._plt_dir is None:
            return

        for dsname in self._l_dsname:
            mod_mm, mod_ee = self._d_model[dsname]
            dat_mm, dat_ee = self._d_data [dsname]

            self._plot(dat_mm, mod_mm, None, f'mm_{dsname}_stk', stacked=True)
            self._plot(dat_ee, mod_ee, None, f'ee_{dsname}_stk', stacked=True)

            self._plot(dat_mm, mod_mm, None, f'mm_{dsname}_ovr', stacked=False)
            self._plot(dat_ee, mod_ee, None, f'ee_{dsname}_ovr', stacked=False)
    #--------------------------------
    def _combine_matrices(self, mat_1, mat_2):
        rows = mat_1.shape[0] + mat_2.shape[0]
        cols = mat_1.shape[1] + mat_2.shape[1]
        
        mat_3 = numpy.zeros((rows, cols))
        
        mat_3[:mat_1.shape[0], :mat_1.shape[1]] = mat_1
        mat_3[mat_1.shape[0]:, mat_1.shape[1]:] = mat_2

        return mat_3
    #--------------------------------
    def _merge_constraints(self, ex_cns, ck_cns):
        if   ex_cns is None: 
            l_par = ck_cns[0]
            l_val = ck_cns[1]
            cov   = ck_cns[2]
        elif ck_cns is None:
            l_par = ex_cns[0]
            l_val = ex_cns[1]
            cov   = ex_cns[2]
        else:
            l_par = ex_cns[0] + ck_cns[0]
            l_val = ex_cns[1] + ck_cns[1]
            cov   = self._combine_matrices(ex_cns[2], ck_cns[2])

        if self._drop_corr:
            log.warning('Using covariance matrix with correlations removed')
            cov_no_diag=numpy.matrix.copy(cov)
            numpy.fill_diagonal(cov_no_diag, 0)
            cov = cov - cov_no_diag

        return l_par, l_val, cov
    #--------------------------------
    def _build_const(self):
        ex_cns = self._get_ex_const()
        ck_cns = self._get_ck_const()

        if ex_cns is None and ck_cns is None:
            return

        l_par, l_val, cov = self._merge_constraints(ex_cns, ck_cns)

        cns   = zfit.constraint.GaussianConstraint(params     = l_par,
                                                   observation= l_val,
                                                   uncertainty= cov)

        self._dump_constraints(l_par, l_val, cov)

        return cns
    #--------------------------------
    def _dump_constraints(self, l_par, l_val, cov):
        npar   = len(l_par)
        l_name = [par.name           for par   in l_par]
        l_cov  = [ cov[i_par][i_par] for i_par in range(npar)] 
        l_err  = [ math.sqrt(cov)    for cov   in l_cov ]
        df     = pnd.DataFrame({'Name' : l_name, 'Value' : l_val, 'Error' : l_err})

        df.to_json(f'{self._plt_dir}/constraints.json', indent=4)
    #--------------------------------
    @utnr.timeit
    def get_fit_result(self):
        self._initialize()

        log.info('Creating likelihoods')
        nll = None 
        for dsname in self._l_dsname:
            mod_mm, mod_ee = self._d_model[dsname]
            dat_mm, dat_ee = self._d_data [dsname]

            nll_mm = zfit.loss.ExtendedUnbinnedNLL(model=mod_mm, data=dat_mm)
            nll_ee = zfit.loss.ExtendedUnbinnedNLL(model=mod_ee, data=dat_ee)

            if nll is None:
                nll = nll_mm + nll_ee
            else:
                nll+= nll_mm + nll_ee

        cns = self._build_const()
        if cns is not None:
            nll = nll.create_new(constraints=cns)

        mnm = zfit.minimize.Minuit()
        log.info('Running minimization')
        res = mnm.minimize(nll)

        self._plot_fit()

        log.info('Finalizing')
        self._finalize()

        return res 
#--------------------------------

